/********

	quickvan.c --- Quick-VAN protocol program for UNIX host
	Automatic upload/download mode for Sterm version 2.00
				07/11/94  Masayuki Suzuki

*********/

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <utime.h>
#include <signal.h>

#include	<errno.h>
#include	<fcntl.h>
#include	<strings.h>
#include	<signal.h>

#include	<sys/termio.h>
#include	<time.h>

#include	<stropts.h>
#include	<sys/conf.h>

/* #define	MESSAGE */

/* If defined AUTO, use automatic start Quick-VAN protocol */
/* This mode is able to use Sterm only */
#define	AUTO

/* Sun OS 4.1.x, JLE 1.1.x for SPARC station/2 */
/* If Solaris 2.x, use /usr/ucb/cc */
/* #define	SUN */

/* If defined DS90, for Fujitsu DS/90 7000 series */
/* #define	DS90 */

#define	CCHMAXPATHCOMP	256
#define	UCHAR	unsigned char
#define	ULONG	unsigned long

#define	RS_TIMES	10

#define	BLOCKSIZE	128
#define	MAX_RETRY	10
#define	WAIT_RECV	20
#define	WAIT_CANCEL	60
#define	WAIT_NEGO	120
#define	WAIT_DATA	200
#define	WAIT_RESPONSE	20

#define	MAX_WINDOW	30

#define	SOH	0x01
#define	STX	0x02
#define	EOT	0x04
#define	ACK	0x06
#define	CR	0x0D
#define	DLE	0x10
#define	NAK	0x15
#define	CAN	0x18
#define	ESC	0x1B
#define	SUM	0x1C

#define	US	0x1F

#define	SINIT	'0'
#define	VFILE	'1'
#define	VENQ	'2'
#define	START	'3'

#define	QUICK	'Q'
#define	RINIT	'R'
#define	VRPOS	'P'
#define	VACK	'A'
#define	VNAK	'N'
#define	VSTAT	'T'
#define	VSKIP	'S'

#define	XFUNC	0
#define	XRBLK	1
#define	XVDAT	2
#define	XCTRL	3

char	msg_esc[]	= "Terminated by ESC key.";
char	msg_del[]	= "Terminated by line error.";
char	msg_over[]	= "Terminated by retry count over.";
char	msg_can[]	= "Terminated by cancel received.";
char	msg_time[]	= "Terminated by timeout.";
char	msg_ack[]	= "Unable to detect ACK.";
char	msg_packet[]	= "Memory allocation error for packet.";
char	msg_retry[]	= "Detected CheckSum error (retry).\r\n";
char	msg_ignore[]	= "Detected CheckSum error (ignore).\r\n";

void	send_func_block(),send_response_block(),send_vdat_block();
void	cancel_quickvan(),PutFileInfo(),TransferMessage();
void	qv_send(),qv_receive(),put_232c();

UCHAR	filename[CCHMAXPATHCOMP+1];
long	fileinfo[6],StartTime;
int	window,resume;
FILE	*fp,*setup_file();
int	a_value,g_value;

#ifdef	MESSAGE
char	debug[128];
#endif

struct	PACKET {
	UCHAR	CTRL;
	UCHAR	TYPE;
	UCHAR	SUB;
	UCHAR	BODY[BLOCKSIZE];
	UCHAR	CHKSUM;
	};

struct	BODY_INIT {
	UCHAR	VerInfo;
	UCHAR	RetryMethod;
	UCHAR	Window[2];
	UCHAR	OptionNumber;
/*	UCHAR	OptionFunc;             ..... unused PC-VAN */
	};

struct	PACKET	*packet_base;

struct termio termio, save_term;
char	tmp_file[32];
int	fd, wd, text;

main(argc,argv)
int	argc;
char	*argv[];
{
static	char	dummy[2] = { 0x00, 0x00 };
struct	stat	sb;
char	*file,*tmp;
int	i,k,opt;

	text = 0;
	opt = -1;
	for(k=1;k<argc;++k) {
		if(strcmp(argv[k],"-t")==0) {
			text = 1;
			strcpy(tmp_file,"/tmp/QuickVAN.XXXXXX");
			mktemp(tmp_file);
		} else if(strcmp(argv[k],"-u")==0) opt = 0;
		else if(strcmp(argv[k],"-d")==0) opt = 1;
		else break;
	}

	if(opt==0) {
		resume = 0;
#ifdef	AUTO
		if(argc<=k) exit(1);
			tmp = argv[k] + strlen(argv[k]);
			while(--tmp>argv[k]) {
				if(*tmp=='/' || *tmp=='\\' || *tmp==':') {
					++tmp;
					break;
				}
			}
#else
		if(argc>k) {
			tmp = argv[k];
#endif
			if(stat(tmp,&sb)==0) {
				if(text) printf("%s: already exist, delete or cancel (d/c) ? ",tmp);
				else printf("%s: already exist, recover, delete or cancel (r/d/c) ? ",tmp);
				while (1) {
					i = getchar();
					if(i=='c') exit(1);
					else if(i=='d') {
						unlink(argv[k]);
						break;
					} else if(!text && i=='r') {
						resume = 1;
						break;
					} else {
						while(getchar()!='\n');
						if(text) printf("(d/c) ? ");
						else printf("(r/d/c) ? ");
					}
				}
			}
			file = argv[k];
#ifndef	AUTO
		} else file = dummy;
		printf("* Please start Quick-VAN Protocol (Upload).\r\n");
#endif
	        if(InitProc()) exit(1);
		qv_receive(file);
	} else if(opt==1) {
		if(stat(argv[k],&sb)) {
			printf("%s: not exist.\r\n",argv[k]);
			exit(1);
		}
#ifndef	AUTO
		printf("* Please start Quick-VAN Protocol (Download).\r\n");
#endif
	        if(InitProc()) exit(1);
		qv_send(argv[k]);
	} else exit(1);
        EndProc();
	exit(0);
}

void qv_send(fname)
char   *fname;
{
int	i,j,k,f,u,v,w,dc,g;
int	rtn_code,StartBlock,size[MAX_WINDOW];
long	mx,ms;
UCHAR	seq,s,buf[128];
struct	PACKET *packet[MAX_WINDOW],*send_packet,*recv_packet;

	put_msg("* * * * * *\r\nQuick-VAN / Download from HOST Procedure Start .....");

	if(text) {
		if(convert_text(0,fname)) goto error;
	}

	packet_base = (struct PACKET *)malloc(sizeof(struct PACKET)*(MAX_WINDOW+2));
#ifdef	AUTO
	if(!packet_base) {
		put_msg(msg_packet);
		fprintf(stderr,"quickvan: %s\r\r\n",msg_packet);
		goto error;
	}
	if(StartProc(0,fname,packet_base)) goto terminate;
#else
	if(!packet_base) {
		cancel_quickvan(2);
		put_msg(msg_packet);
		goto error;
	}
#endif
	if(SendHandShake(packet_base)) goto terminate;

	strcpy(filename,fname);
	fp = setup_file(1,filename,&i);
	if(fp==(FILE *)NULL) {
		cancel_quickvan(0);
		goto terminate;
	}
	PutFileInfo();

	recv_packet = packet_base;
	send_packet = packet_base + 1;

	StartBlock = GetStartBlock(send_packet);
	if(StartBlock<0) {
		fclose(fp);
		goto terminate;
	}

	if(StartBlock) {
		ms = (long)BLOCKSIZE * (long)StartBlock;
		if(fseek(fp,ms,0)) {
			fclose(fp);
			put_str(filename);
			put_str(": seek error for recover transfer.\r\n");
			cancel_quickvan(2);
			goto terminate;
		}
		fileinfo[0] -= (long)StartBlock * (long)BLOCKSIZE;
	}

	for(i=0;i<window;++i) {
		packet[i] = packet_base + i + 2;
		packet[i]->CTRL = 0x00;
	}

	u = v = w = g = 0;
	mx = 0L;
	StartTime = time(0L);
	for(dc=0,f=1,seq=0;;) {
		if(f>0 && g<g_value) {
			k = fread(packet[w]->BODY,1,BLOCKSIZE,fp);
			if(k>0) {
				++seq;
				if(k<BLOCKSIZE) memset(&packet[w]->BODY[k],0x00,BLOCKSIZE-k);
				packet[w]->CTRL = SOH;
				packet[w]->TYPE = seq;
				packet[w]->SUB = ~seq;
				size[w] = k;
				++w;
				++g;
				if(w>=window) w = 0;
			} else {
				if(feof(fp)) f = 0;
				else {
					fclose(fp);
					cancel_quickvan(2);
					put_chr('\n');
					put_msg(" Read error ");
					goto terminate;
				}
			}
		}
		if(v!=w && packet[v]->CTRL==SOH) {
			send_vdat_block(packet[v]);
			mx += size[v];
			++v;
			if(v>=window) v = 0;
			rtn_code = receive(0,recv_packet);
			if(rtn_code<0) continue;
		} else {
			if(f==0 && v==w) {
				memset(send_packet->BODY,0x00,BLOCKSIZE);
				send_packet->CTRL = SOH;
				send_packet->TYPE = 0x00;
				send_packet->SUB = VENQ;
				send_packet->BODY[0] = seq;
				send_packet->BODY[1] = '0';
				send_func_block(send_packet);
				f = -1;
			}
			rtn_code = receive(WAIT_DATA,recv_packet);
			if(rtn_code<0) {
				fclose(fp);
				cancel_quickvan(2);
				put_chr('\n');
				put_msg(msg_time);
				goto terminate;
			}
		}
		if(recv_packet->CTRL==CAN) {
			fclose(fp);
			put_232c(EOT);
			sleep(1);
			put_chr('\n');
			put_msg(msg_can);
			goto terminate;
		} else if(recv_packet->CTRL==SUM) {
			put_chr('\n');
			put_str(msg_ignore);
		} else if(recv_packet->CTRL==ESC) {
			fclose(fp);
			cancel_quickvan(2);
			put_chr('\n');
			put_msg(msg_esc);
			goto terminate;
		} else if(rtn_code==XRBLK) {
			if(recv_packet->TYPE==VACK) {
				for(i=0;i<window;++i) {
					if(packet[i]->CTRL==SOH &&
						recv_packet->SUB==(packet[i]->TYPE|0x80)) break;
				}
				if(i<window) {
					if(i<u) {
						for(;u<window;++u) {
							packet[u]->CTRL = 0x00;
							--g;
						}
						u = 0;
					}
					for(;u<=i;++u) {
						packet[u]->CTRL = 0x00;
						--g;
					}
				}
			} else if(recv_packet->TYPE==VNAK) {
				for(i=0;i<window;++i) {
					if(packet[i]->CTRL==SOH &&
						recv_packet->SUB==(packet[i]->TYPE|0x80)) break;
				}
				if(i<window) {
					if(i<u) {
						for(;u<window;++u) {
							packet[u]->CTRL = 0x00;
							--g;
						}
						u = 0;
					}
					for(;u<i;++u) {
						packet[u]->CTRL = 0x00;
						--g;
					}
					if(u>v) {
						for(j=u;j<window;++j) mx -= size[j];
						j = 0;
					} else j = u;
					for(;j<v;++j) mx -= size[j];
					v = u;
					if(f<0) f = 0;
				}
			} else if(recv_packet->TYPE==VSTAT) {
				if(recv_packet->SUB==(seq|0x80)
					&& recv_packet->BODY[0]=='0') break;
			}
		}
	}

	fclose(fp);
	TransferMessage();

	put_232c(EOT);
	while (1) {
		rtn_code = receive(WAIT_DATA,recv_packet);
		if(rtn_code<0) {
			cancel_quickvan(2);
			put_msg(msg_ack);
			break;
		} else if(recv_packet->CTRL==ACK) break;
	}

terminate:
	free(packet_base);

error:
	if(text) unlink(tmp_file);
	put_msg("Quick-VAN / Download from HOST completed !\r\n* * * * * *");

	return;
}

static void PutFileInfo() {
char	buf[128];

	sprintf(buf,"%ld bytes\r\n",fileinfo[0]);
	put_str("File name: ");
	put_str(filename);
	put_str("\r\nFile size: ");
	put_str(buf);
	return;
}

static void TransferMessage() {
int	i,j,k;
long	tv,mx;
char	buf[1280];

	tv = time(0L) - StartTime + 1;
	i = tv / 3600L;
	j = tv % 3600L;
	k = j % 60;
	j /= 60;
	mx = fileinfo[0] / tv;
	sprintf(buf,"\r\n>> Execution time: %02d:%02d:%02d\r\n",i,j,k);
	put_str(buf);
	sprintf(buf,  ">> Transfer rate: %ld bytes/sec\r\n",mx);
	put_str(buf);
	return;
}

SendHandShake(packet)
struct	PACKET	*packet;
{
int	rtn_code,w;

retry:
	memset(packet->BODY,0x00,BLOCKSIZE);
	packet->SUB = SINIT;
	sprintf((char *)packet->BODY,"10%02d00",MAX_WINDOW);
	send_func_block(packet);

	while (1) {
		rtn_code = receive(WAIT_NEGO,packet);
		if(rtn_code<0) {
			put_msg(msg_time);
			goto cancel;
		} else if(packet->CTRL==NAK) {
			goto retry;
		} else if(packet->CTRL==CAN) {
			put_msg(msg_can);
			return(-1);
		} else if(packet->CTRL==ESC) {
			put_msg(msg_esc);
			goto cancel;
		} else if(rtn_code==XRBLK && packet->TYPE==RINIT && packet->SUB==0x80) {
			if(packet->BODY[0]!='1') goto cancel;
			if(packet->BODY[1]!='0') goto cancel;
			w = (packet->BODY[2]-'0')*10 + packet->BODY[3] - '0';
			if(w>MAX_WINDOW) {
				put_msg("exceeded window max");
				goto cancel;
			}
			window = w;
			if(packet->BODY[4]!='0') goto cancel;
			if(packet->BODY[5]!='0' && packet->BODY[5]!=0x00) goto cancel;
			a_value = w / 3;
			g_value = w - 1;
			if(a_value<1) a_value = 1;
			break;
		}
	}

	return(0);

cancel:
	cancel_quickvan(2);
	return(-1);
}

GetStartBlock(packet)
struct	PACKET	*packet;
{
int	i,j,k,m,sp;
int	rtn_code;
char	buf[CCHMAXPATHCOMP+1];
void	SetTransferName();

	memset(packet->BODY,0x00,BLOCKSIZE);
	packet->SUB = VFILE;
	SetTransferName(buf);
	i = fileinfo[5] / 3600L;
	j = fileinfo[5] % 3600L;
	k = j % 60;
	j /= 60;
	sprintf((char *)packet->BODY,"01%s%c%ld%c%08ld%02d%02d%02d",
				buf,0,fileinfo[0],0,fileinfo[4],i,j,k);
	send_func_block(packet);

	while (1) {
		rtn_code = receive(WAIT_NEGO,packet);
		if(rtn_code<0) {
			put_msg(msg_time);
			goto cancel;
		} else if(packet->CTRL==CAN) {
			put_msg(msg_can);
			return(-1);
		} else if(packet->CTRL==ESC) {
			put_msg(msg_esc);
			goto cancel;
		} else if(rtn_code==XRBLK && packet->TYPE==VRPOS && packet->SUB==0x80) {
			sp = atoi((char *)packet->BODY);
			break;
		}
	}

	return(sp);

cancel:
	cancel_quickvan(2);
	return(-1);
}

static void SetTransferName(buf)
char *buf;
{
int	i,j,k,m;

	for(m=strlen(filename)-1;m>=0;--m) {
		if(filename[m]=='/' || filename[m]=='\\' || filename[m]==':') break;
	}
	strcpy(buf,&filename[++m]);
	return;
}

void qv_receive(fname)
char   *fname;
{
int	i,j,k,m,n,ack;
int	rtn_code,StartBlock;
long	mx,ms;
UCHAR	seq,s,buf[128];
void	analize_fileinfo();
struct	PACKET *packet,*send_packet,*recv_packet;

	put_msg("* * * * * *\r\nQuick-VAN / Upload to HOST Procedure Start .....");

	packet_base = (struct PACKET *)malloc(sizeof(struct PACKET)*3);
#ifdef	AUTO
	if(!packet_base) {
		put_msg(msg_packet);
		fprintf(stderr,"quickvan: %s\r\n",msg_packet);
		goto error;
	}
	if(StartProc(1,fname,packet_base)) goto terminate;
#else
	if(!packet_base) {
		cancel_quickvan(2);
		put_msg(msg_packet);
		goto error;
	}
#endif
	if(ReceiveHandShake(packet_base)) goto terminate;

	recv_packet = packet_base;
	send_packet = packet_base + 1;
	packet = packet_base + 2;

	for (j=0;;++j) {
		for (i=0;;) {
			rtn_code = receive(WAIT_RECV,recv_packet);
			if(rtn_code==XFUNC && recv_packet->CTRL==SOH && recv_packet->SUB==VFILE) break;
			else if(recv_packet->CTRL==EOT) {
				put_232c(ACK);
				sleep(1);
				goto terminate;
			} else if(recv_packet->CTRL==CAN) {
				put_msg(msg_can);
				goto terminate;
			} else if(recv_packet->CTRL==DLE) {
				put_msg(msg_del);
				goto terminate;
			} else if(recv_packet->CTRL==ESC) {
				put_msg(msg_esc);
				cancel_quickvan(0);
				goto terminate;
			} else if(rtn_code<0 || (rtn_code==XFUNC && recv_packet->CTRL==SUM)) {
				if(recv_packet->CTRL==SUM) {
					put_msg(msg_retry);
				}
				++i;
				if(i>=MAX_RETRY) {
					put_msg(msg_over);
					cancel_quickvan(0);
					goto terminate;
				} else {
					if(j) send_response_block(send_packet,1);
					else  send_response_block(send_packet,sizeof(struct BODY_INIT));
				}
			}
		}

		analize_fileinfo(recv_packet->BODY,fname);
		PutFileInfo();
		fp = setup_file(0,filename,&StartBlock);
		if(fp==(FILE *)NULL) {
			cancel_quickvan(0);
			goto terminate;
		}

		send_packet->TYPE = VRPOS;
		send_packet->SUB  = 0x80;
		if(StartBlock) {
			sprintf((char *)send_packet->BODY,"%d",StartBlock);
			fileinfo[0] -= (long)StartBlock * (long)BLOCKSIZE;
		} else send_packet->BODY[0] = 0x00;
		send_response_block(send_packet,strlen(send_packet->BODY));

		mx = 0L;
		StartTime = time(0L);
		for (i=0,n=0,ack=0,seq=1;;) {
			rtn_code = receive(WAIT_RECV,recv_packet);
			if(rtn_code==XVDAT && recv_packet->CTRL!=SUM) {
				if(recv_packet->TYPE==seq) {
					ms = fileinfo[0] - mx;
					if(ms<=0 || ms>BLOCKSIZE) ms = BLOCKSIZE;
					if(fwrite(recv_packet->BODY,1,(size_t)ms,fp)<(size_t)ms) {
						fclose(fp);
						put_chr('\n');
						put_msg(" Write error ");
						cancel_quickvan(1);
						goto terminate;
					}
					mx += ms;
					if(++ack>=a_value) {
						ack = 0;
						packet->TYPE = VACK;
						packet->SUB = (seq&0x7F) | 0x80;
						send_response_block(packet,0);
					}
					++seq;
				}
				n = 0;
				++i;
			} else if(rtn_code==XFUNC && recv_packet->SUB==VENQ) {
				s = recv_packet->BODY[0]+1;
				if(seq==s) break;
			} else if(recv_packet->CTRL==CAN) {
				fclose(fp);
				put_chr('\n');
				put_msg(msg_can);
				goto terminate;
			} else if(recv_packet->CTRL==ESC) {
				fclose(fp);
				put_chr('\n');
				put_msg(msg_esc);
				cancel_quickvan(1);
				goto terminate;
			} else if(recv_packet->CTRL==DLE) {
				fclose(fp);
				put_chr('\n');
				put_msg(msg_del);
				goto terminate;
			} else if(rtn_code<0 || (rtn_code==XVDAT && recv_packet->CTRL==SUM)) {
				++n;
				if(n>=MAX_RETRY) {
					fclose(fp);
					put_chr('\n');
					put_msg(msg_over);
					cancel_quickvan(1);
					goto terminate;
				}
				if(i || rtn_code==XVDAT) {
					ack = 0;
					packet->TYPE = VNAK;
					packet->SUB = (seq&0x7F) | 0x80;
					send_response_block(packet,0);
				} else  send_response_block(send_packet,strlen(send_packet->BODY));
			}
		}

		fclose(fp);
		set_file_time(filename,fileinfo);

		TransferMessage();

		send_packet->TYPE = VSTAT;
		send_packet->SUB  = recv_packet->BODY[0] | 0x80;
		send_packet->BODY[0] = '0';
		send_response_block(send_packet,1);
	}

terminate:
	free(packet_base);

error:
	if(text) {
		convert_text(1,filename);
		unlink(tmp_file);
	}
	put_msg("Quick-VAN / Upload to HOST completed !\r\n* * * * * *");

	return;
}

static ReceiveHandShake(packet)
struct	PACKET	*packet;
{
int	i,j,k;
int	rtn_code;
void	send_rinit();

	put_232c(NAK);
	put_str("send [NAK]\r\n");
	for(i=0;;) {
		rtn_code = receive(WAIT_RECV,packet);
		if(rtn_code==XFUNC && packet->CTRL==SOH && packet->SUB==SINIT) {
			send_rinit(packet);
			return(0);
		} else if(packet->CTRL==CAN) {
			put_msg(msg_can);
			return(-1);
		} else if(packet->CTRL==DLE) {
			put_msg(msg_del);
			return(-1);
		} else if(packet->CTRL==ESC) {
			put_msg(msg_esc);
			cancel_quickvan(0);
			return(-1);
		} else if(rtn_code<0 || (rtn_code==XFUNC && packet->CTRL==SUM)) {
			if(packet->CTRL==SUM) put_msg(msg_ignore);
			if(++i>=MAX_RETRY) break;
			put_232c(NAK);
			put_str("send [NAK]\r\n");
		}
	}
	put_msg(msg_over);
	cancel_quickvan(0);
	return(-1);
}

static FILE *setup_file(opt,name,stblk)
int	opt,*stblk;
char	*name;
{
int	i,j,k;
long	tmpinfo[6];
FILE	*fx;
struct stat buf;
char	*tmp;

	if(text) name = tmp_file;

	if(opt) {
		j = stat(name,&buf);
		if(j!=0) {
			put_str(name);
			put_str(": not found.\r\n");
			return((FILE *)NULL);
		}
		
		if(get_file_information(name,fileinfo)) {
			put_str(name);
			put_str(": unable to get file information.\r\n");
			return((FILE *)NULL);
		}
		
		fx = fopen(name,"r");
		if(fx==NULL) {
			put_str(name);
			put_str(": can't open for read.\r\n");
		}
	} else {
#ifdef	AUTO
		if(!text) {
			tmp = name;
			name = tmp + strlen(tmp);
			while ( --name>tmp ) {
				if(*name=='/' || *name=='\\' || *name==':') {
					++name;
					break;
				}
			}
		}
#endif
		j = stat(name,&buf);
		if(j==0) {
			if(resume) {
				if(get_file_information(name,tmpinfo)) {
					put_str(name);
					put_str(": unable to get file information.\r\n");
					return((FILE *)NULL);
				}
				tmpinfo[1] = tmpinfo[0] / (long)BLOCKSIZE;
				tmpinfo[2] = tmpinfo[1] * (long)BLOCKSIZE;
				fx = fopen(name,"r+");
				if(fx==(FILE *)NULL) {
					put_str(name);
					put_str(": can't open for append.\r\n");
					return((FILE *)NULL);
				}
				if(fseek(fx,tmpinfo[2],0)) {
					put_str(name);
					put_str(": can't seek for append.\r\n");
					return((FILE *)NULL);
				}
				*stblk = tmpinfo[1];
				return(fx);
			} else {
				put_str(name);
				put_str(": already exist, deleted.\r\n");
				unlink(name);
			}
		}
		fx = fopen(name,"w");
		if(fx==(FILE *)NULL) {
			put_str(name);
			put_str(": can't open for write.\r\n");
		}
		*stblk = 0;
	}
	return(fx);
}

static void analize_fileinfo(buf,name)
UCHAR	*buf,*name;
{
int	i,j,k;

	buf += 2;
	strcpy(filename,buf);
	k = strlen(filename) - 4;
	if(k>=0 && strcmp(&filename[k],".###")==0) filename[k] = 0x00;
	if(*name) {
		strcpy(filename,name);
		*name = 0x00;
	}
	buf += (strlen(buf)+1);
	for (fileinfo[0]=0;*buf;) {
		fileinfo[0] = fileinfo[0]*10L + *buf++ - '0';
	}
	++buf;
	fileinfo[1] = -540;
	if(*buf) {
		for(i=0,fileinfo[2]=0;i<8;++i) {
			fileinfo[2] = fileinfo[2]*10L + *(buf+i) - '0';
		}
	} else fileinfo[2] = 0L;
	buf += 8;
	if(*buf) {
		fileinfo[3] = (long)(*buf-'0') * 36000L
				+ (long)(*(buf+1)-'0') * 3600L
				+ (long)(*(buf+2)-'0') * 600L
				+ (long)(*(buf+3)-'0') * 60L
				+ (long)(*(buf+4)-'0') * 10L
				+ (long)(*(buf+5)-'0');
	} else fileinfo[3] = 0L;
	fileinfo[4] = fileinfo[2];
	fileinfo[5] = fileinfo[3];
	return;
}

static void send_rinit(receive_packet)
struct	PACKET	*receive_packet;
{
struct	BODY_INIT	*sinit,*rinit;
struct	PACKET		*send_packet;
int	w;

	sinit = (struct BODY_INIT *)receive_packet->BODY;
	send_packet = receive_packet + 1;
	rinit = (struct BODY_INIT *)send_packet->BODY;

	send_packet->TYPE = RINIT;
	send_packet->SUB  = 0x80;
	rinit->VerInfo = '1';
	rinit->RetryMethod = '0';
	w = (sinit->Window[0]-'0')*10 + sinit->Window[1]-'0';
	if(w>MAX_WINDOW) w = MAX_WINDOW;
	sprintf((char *)rinit->Window,"%02d",w);
	rinit->OptionNumber = '0';
/*	rinit->OptionFunc = '0'; */
	window = w;
	a_value = w / 3;
        g_value = w - 1;
	if(a_value<1) a_value = 1;

	send_response_block(send_packet,sizeof(struct BODY_INIT));

	return;
}

static void send_func_block(packet)
struct	PACKET	*packet;
{
int	k;
UCHAR	sum;

	sum = 0;
	packet->CTRL = SOH;
	put_232c(packet->CTRL);
	sum += packet->CTRL;
	packet->TYPE = 0x00;
	put_232c(packet->TYPE);
	sum += packet->TYPE;
	put_232c(packet->SUB);
	sum += packet->SUB;
	for(k=0;k<BLOCKSIZE;++k) {
		put_232c(packet->BODY[k]);
		sum += packet->BODY[k];
	}
	put_232c(sum);
	return;
}

static void send_response_block(packet,size)
struct	PACKET	*packet;
int		size;
{
int	k;
UCHAR	sum;

	sum = 0;
	packet->CTRL = STX;
	put_232c(packet->CTRL);
	sum += packet->CTRL;
	put_232c(packet->TYPE);
	sum += packet->TYPE;
	put_232c(packet->SUB);
	sum += packet->SUB;
	for(k=0;k<size;++k) {
		put_232c(packet->BODY[k]);
		sum += packet->BODY[k];
	}
	sum |= 0x80;
	put_232c(sum);
	put_232c(CR);
	return;
}

static void send_vdat_block(packet)
struct	PACKET	*packet;
{
int	k;
UCHAR	sum;

	sum = 0;
	put_232c(packet->CTRL);
	put_232c(packet->TYPE);
	put_232c(packet->SUB);
	for(k=0;k<BLOCKSIZE;++k) {
		put_232c(packet->BODY[k]);
		sum += packet->BODY[k];
	}
	put_232c(sum);
	return;
}

static void cancel_quickvan(opt)
int	opt;
{
int	rtn_code;

	put_232c(CAN);
	if(opt==1) {
		while (1) {
			rtn_code = receive(WAIT_CANCEL,packet_base);
			if(packet_base->CTRL==EOT || rtn_code<0) break;
		}
	} else if(opt==2) {
		while(receive(1,packet_base)>=0);
	} else {
		sleep(1);
	}
	return;
}

static receive(wait,buf)
int	wait;
struct	PACKET	*buf;
{
int	i,j,k,kx;
int	t,vt,type;
UCHAR	sum;

	if(wait) {
		t = wait * RS_TIMES;
		vt = 1;
	} else {
		t = 1;
		vt = 0;
	}

	buf->CTRL = 0x00;
	for(i=0;i<t;) {
              	k = get_232c(vt);
              	if(k<0) {
              		if(k==-2) {
              			buf->CTRL = DLE;
              			return(XCTRL);
              		}
              		++i;
                        continue;
              	} else if(vt==0) {
              		vt = 1;
              		t = WAIT_RESPONSE;
              	}

retry:
		if(buf->CTRL==0x00) {
			switch (k) {
				case SOH:
					buf->CTRL = SOH;
					sum = 0;
					j = -2;
					break;
				case STX:
					buf->CTRL = STX;
					sum = STX;
					j = -2;
					break;
				case EOT:
				case ACK:
				case NAK:
				case CAN:
	              			buf->CTRL = k;
					k = get_232c(RS_TIMES);
					if(k<0 || (k==0x0D && (buf->CTRL==EOT || buf->CTRL==CAN))) {
              					return(XCTRL);
					} else {
						buf->CTRL = 0x00;
						goto retry;
					}
			}
		} else if(buf->CTRL==SOH) {
			if(j==-2) buf->TYPE = k;
			else if(j==-1) {
				buf->SUB = k;
				if(buf->TYPE==0x00 && buf->SUB>=SINIT
#ifdef	AUTO
							 && buf->SUB<=START) {
#else
							 && buf->SUB<=VENQ) {
#endif
					sum += buf->CTRL;
					sum += buf->TYPE;
					sum += buf->SUB;
					type = XFUNC;
				} else if(buf->TYPE==(UCHAR)~buf->SUB) {
					type = XVDAT;
				} else {
					buf->CTRL = 0x00;
					continue;			
				}
			} else if(j<BLOCKSIZE) {
				buf->BODY[j] = k;
				sum += k;
			} else {
				buf->CHKSUM = k;
				if(buf->CHKSUM!=sum) {
					buf->CTRL = SUM;
				}
				return(type);
			}
			++j;
		} else if(buf->CTRL==STX) {
			if(j==-2) {
				if(k!=RINIT && k!=VRPOS && k!=VACK &&
					k!=VNAK && k!=VSTAT && k!=VSKIP) {
					buf->CTRL = 0x00;
					continue;
				} else buf->TYPE = k;
			} else if(j==-1) buf->SUB = k;
			else if(k!='\r') {
				if(j>=(BLOCKSIZE-3)) {
					buf->CTRL = 0x00;
					continue;
				}
				buf->BODY[j] = k;
			} else {
				buf->CHKSUM = buf->BODY[j-1];
				buf->BODY[j-1] = 0x00;
				sum -= buf->CHKSUM;
				sum |= 0x80;
				if(sum!=buf->CHKSUM) {
					buf->CTRL = SUM;
				}
				return(XRBLK);
			}
			sum += k;
			++j;
		} else {
			buf->CTRL = 0x00;
			goto retry;
		}
	}

	buf->CTRL = 0x00;
	return(-1);
}

#ifdef	AUTO
static StartProc(opt,fname,packet)
int	opt;
char	*fname;
struct	PACKET	*packet;
{
int	i,j,k;
int	rtn_code;
void	send_automatic();

	put_232c(US);
	put_str("send [US]\r\n");
	for(i=0;;) {
		rtn_code = receive(WAIT_RECV,packet);
		if(rtn_code==XFUNC && packet->SUB==START) {
			if(packet->CTRL==SUM) {
				put_str(msg_ignore);
			} else {
				send_automatic(opt,fname,packet);
				if(opt) sleep(1);
				return(0);
			}
		} else if(packet->CTRL==CAN) {
			put_msg(msg_can);
			return(-1);
		} else if(packet->CTRL==DLE) {
			put_msg(msg_del);
			return(-1);
		} else if(packet->CTRL==ESC) {
			put_msg(msg_esc);
			cancel_quickvan(0);
			return(-1);
		} else if(rtn_code<0) {
			if(++i>=MAX_RETRY) break;
			put_232c(US);
			put_str("send [US]\r\n");
		}
	}
	put_msg(msg_over);
	cancel_quickvan(0);
	return(-1);
}

static void send_automatic(opt,fname,receive_packet)
int	opt;
char	*fname;
struct	PACKET	*receive_packet;
{
struct	PACKET		*send_packet;
int	w;

	w = 0;
	send_packet = receive_packet + 1;

	send_packet->TYPE = QUICK;
	send_packet->SUB  = 0x80;
	sprintf((char *)send_packet->BODY,"%d%d%d%d%s",opt,text,w,w,fname);
	send_response_block(send_packet,strlen(send_packet->BODY));

	return;
}

#endif

get_file_information(filename,info)
char *filename;
long *info;
{
	struct stat buf;
	int sts;

	if (stat(filename, &buf) == 0) {
		struct tm *lt;

		info[0] = buf.st_size;
		info[1] = (ULONG) -540;
		lt = localtime((time_t *) &(buf.st_ctime));
		info[2] = ((ULONG)lt->tm_year+1900L)*10000L +
			((ULONG)lt->tm_mon+1)*100L + (ULONG)lt->tm_mday;
		info[3] = (ULONG)lt->tm_hour*3600L +
				(ULONG)lt->tm_min * 60L + (ULONG)lt->tm_sec;
		lt = localtime((time_t *) &(buf.st_mtime));
		info[4] = ((ULONG)lt->tm_year+1900L)*10000L +
			((ULONG)lt->tm_mon+1)*100L + (ULONG)lt->tm_mday;
		info[5] = (ULONG)lt->tm_hour*3600L +
				(ULONG)lt->tm_min * 60L + (ULONG)lt->tm_sec;
		sts = 0;
	} else {
		sts = -1;
	}
	return sts;
}

set_file_time(file,info)
char *file;
long *info;
{
	struct tm t;
	struct utimbuf tp;

	if(info[2]) {
		t.tm_sec = (int) (info[3] % 60);
		t.tm_min = (int) ((info[3] / 60) % 60);
		t.tm_hour = (int) (info[3] / 3600);
		t.tm_mday = info[2] % 100;
		t.tm_mon = (info[2] / 100) % 100 - 1;
		t.tm_year = (info[2] / 10000) - 1900;
#ifdef	DS90
		tp.actime = mktime(&t);
#else
		tp.actime = timelocal(&t);
#endif
	} else  tp.actime = time(0L);

	if(info[4]) {
		t.tm_sec = (int) (info[5] % 60);
		t.tm_min = (int) ((info[5] / 60) % 60);
		t.tm_hour = (int) (info[5] / 3600);
		t.tm_mday = info[4] % 100;
		t.tm_mon = (info[4] / 100) % 100 - 1;
		t.tm_year = (info[4] / 10000) - 1900;
#ifdef	DS90
		tp.modtime = mktime(&t);
#else
		tp.modtime = timelocal(&t);
#endif
	} else	tp.modtime = time(0L);

	return utime(file, &tp);
}

#ifdef	MESSAGE
FILE *console;
#endif
InitProc()
{
	char	*p, *ttyname();
	int	i;

#ifdef	MESSAGE
	console = fopen("/dev/console","w");
#endif

	p = ttyname(fileno(stdin));
	if ((fd = open(p, O_RDWR | O_NDELAY)) < 0) {
		return(-1);
	}

	wd = fd;

	if (ChangeIO() != 0) {
		ioctl(fd, TCSETA, &save_term);
		return(-1);
	}

	return(0);
}

ChangeIO()
{
extern	void xquit();

	if (ioctl(fd, TCGETA, &termio) == -1) {
		put_msg("TCGETA error");
		return(-1);
	}
	memcpy(&save_term, &termio, sizeof(termio));

	if(change_stream(0)) {
		put_msg("change stream error");
		return(-1);
	}

	termio.c_iflag = 0x0005;
	termio.c_oflag = 0x0000;
	termio.c_lflag = 0x0080;
	termio.c_cc[VINTR] = 0x00;
	termio.c_cc[VQUIT] = 0x00;
	termio.c_cc[VERASE]= 0x00;
	termio.c_cc[VKILL] = 0x00;
	termio.c_cc[VMIN]  = 0x00;
	termio.c_cc[VEOL]  = 0x00;
	termio.c_cc[VTIME] = 0x00;

	if (ioctl(fd, TCSETA, &termio) == -1) {
		change_stream(1);
		put_msg("TCSETA error");
		return(-1);
	}

	signal(SIGTERM,xquit);
	signal(SIGTSTP,xquit);

	return(0);
}

EndProc()
{
#ifdef	MESSAGE
	fclose(console);
#endif
	change_stream(1);

	ioctl(fd, TCSETA, &save_term);
	close(fd);
	sleep(1);
	return(0);
}

void xquit() {
	EndProc();
	exit(2);
}

#define	MAX_MODULE	10
change_stream(opt)
int opt;
{
static	char	buf[MAX_MODULE][FMNAMESZ+1];
static	int	module;
int	i,j,k;

	if(opt) {
		ioctl(fd,I_POP,0);
		ioctl(fd,I_POP,0);
		for(--module;module>=0;--module)
			ioctl(fd,I_PUSH,buf[module]);
	} else {
		for(module=0;module<MAX_MODULE;++module) {
			buf[module][0] = 0x00;
			if(ioctl(fd,I_LOOK,buf[module])) {
				if(errno==EINVAL) break;
				else goto error;
			}
			if(ioctl(fd,I_POP,0)) goto error;
		}
		ioctl(fd,I_PUSH,"ldterm");
		ioctl(fd,I_PUSH,"ttcompat");
	}
	return(0);

error:
	for(--module;module>=0;--module)
			ioctl(fd,I_PUSH,buf[module]);
	return(1);
}

#define	RS_SLEEP 100000
			
void put_232c(x)
char x;
{
loop:
	if(write(wd,&x,1)!=1) goto loop;
	return;
}
#define	BUFSIZE	1024
get_232c(q)
int	q;
{
	static unsigned char	buf[BUFSIZE];
	static int count = 0, stat = 0;

	if(count<stat) {
		return(buf[count++]);
	}

	for(;q>=0;--q) {
		stat = read(fd,buf,BUFSIZE);
		if(stat<0) {
			return(-2);
		} else if(stat==0) {
			if(q==0) return(-1);
			else usleep(RS_SLEEP);
		} else {
			count = 0;
			return((unsigned int)buf[count++]);
		}
	}
	return(-3);
}

convert_text(opt,name)
int	opt;
char	*name;
{
FILE	*xp,*yp;
char	*tmp;
int	k,m;
void	euctosj(),sjtoeuc();

	if(opt) {
#ifdef AUTO
		tmp = name;
		name = tmp + strlen(tmp);
		while ( --name>tmp ) {
			if(*name=='/' || *name=='\\' || *name==':') {
				++name;
				break;
			}
		}
#endif
		xp = fopen(tmp_file,"r");
		if(xp==NULL) return(1);
		yp = fopen(name,"w");
		if(yp==NULL) return(1);
		while (1) {
			k = fgetc(xp);
			if(k<0 || k==0x1A) break;
			else if(k==0x0D) {
				m = fgetc(xp);
				if(m!=0x0A) fputc(k,yp);
				fputc(m,yp);
			} else if(k>=0xA0 && k<=0xDF) {
				fputc(0x8E,yp);
				fputc(k,yp);
			} else if(k>=0x80) {
				m = fgetc(xp);
				sjtoeuc(&k,&m);
				fputc(k,yp);
				fputc(m,yp);
			} else fputc(k,yp);
		}
		fclose(xp);
		fclose(yp);
	} else {
		xp = fopen(name,"r");
		if(xp==NULL) return(1);
		yp = fopen(tmp_file,"w");
		if(yp==NULL) return(1);
		while (1) {
			k = fgetc(xp);
			if(k<0) break;
			else if(k==0x0A) {
				fputc(0x0D,yp);
				fputc(k,yp);
			} else if(k==0x8E) {
				m = fgetc(xp);
				fputc(m,yp);
			} else if(k>=0x80) {
				m = fgetc(xp);
				euctosj(&k,&m);
				fputc(k,yp);
				fputc(m,yp);
			} else fputc(k,yp);
		}
		fclose(xp);
		fclose(yp);
	}
	return(0);
}

void sjtoeuc(x,y)
int	*x,*y;
{
   unsigned int    ch,cl;

         ch = *x;
         cl = *y;

         if(cl<0x9F)
              {  if(cl<0x7F) cl -= 0x1F;
                     else    cl -= 0x20;
                 if(ch<0xA0) ch = (ch-0x81)*2 + 0x21;
                     else    ch = (ch-0xE0)*2 + 0x5F; }
            else
              {  cl -= 0x7E;
                 if(ch<0xA0) ch = (ch-0x81)*2 + 0x22;
                     else    ch = (ch-0xE0)*2 + 0x60; }

         *x = ch | 0x80;
         *y = cl | 0x80;

         return;
}

void euctosj(x,y)
int	*x,*y;
{
unsigned int    ch,cl;

         ch = *x & 0x7F;
         cl = *y & 0x7F;

         if((ch%2)==0)
            {
              if(ch<=0x5E) ch = (ch-0x22)/2 + 0x81;
                  else     ch = (ch-0x22)/2 + 0xC1;
              cl += 0x7E;
            }
          else
            {
              if(ch<=0x5E) ch = (ch-0x21)/2 + 0x81;
                  else     ch = (ch-0x21)/2 + 0xC1;
              if(cl<=0x5F) cl += 0x1F;
                  else     cl += 0x20;
            }

         *x = ch;
         *y = cl;

         return;
}

put_msg(x)
char *x;
{
#ifdef MESSAGE
	fprintf(console,"%s\r\n",x);
#endif
	return;
}
put_str(x)
char *x;
{
#ifdef MESSAGE
	fprintf(console,"%s",x);
#endif
	return;
}
put_chr(x)
char x;
{
#ifdef MESSAGE
	if(x=='\n') fputc('\r',console);
	fputc(x,console);
#endif
	return;
}
